/*
 *  Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *  copyright notice, this list of conditions and the following disclaimer
 *  in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *  include, in the documentation and/or other materials provided with the
 *  distribution, a statement that the software has been modified and
 *  identify the person who was responsible for the modifications.
 *  3. Neither the name of the author nor Forward Computing and Control may be
 *  used to endorse or promote products derived from this software
 *  without specific prior written permission.
 *  4.  All advertising materials mentioning features or use of this
 *  software must display the following acknowledgment:
 *  "This product includes software developed by Matthew Ford
 *  and Forward Computing and Control Pty. Ltd.
 *  (http://www.forward.com.au)."
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
/*
 *  java -jar RunWST-1.5.0.jar templateName
 */
/*
 *  java -Djava.util.logging.config.file=runWSTlogging.properties -jar RunWST-1.5.jar templateName
 */
package au.com.forward.runWST;

import java.io.*;
import java.util.*;
import java.util.logging.*;
import au.com.forward.logging.*;
import java.awt.*;
import java.awt.event.*;
import java.nio.channels.FileChannel;

import au.com.forward.runWST.wstAttributeReader.*;
import au.com.forward.runWST.maps.*;
import au.com.forward.runWST.misc.*;
import org.antlr.stringtemplate.*;
import au.com.forward.webstringtemplate.*;
import au.com.forward.webstringtemplate.filters.*;

/**
 *  This is the Web Page Designer's test program for WebStringTemplates
 *
 *@author     matthew ford
 *@created    November 26, 2003
 */
public class RunWST {

	/**
	 *  The Java logger for this class
	 */
	public final static Logger logger = Logger.getLogger("au.com.forward.runWST.RunWST");

	/**
	 *  The version number 1.4
	 */
	public final static String VERSION = "1.4";

	/**
	 *  The version date 15th Jan. 2004
	 */
	public final static String VERSION_DATE = "15th Jan. 2004";

	static {
		String logFileName = "RunWST.log";
		try {
			String loggingConfigFile = System.getProperty("java.util.logging.config.file");
			if (loggingConfigFile != null) {
				// OK Initialize the logging if there is a logging config file specified.
				
				// use the following line to write the log file to the current dir.
				String userHome = "";
				// use the following line to write the log to the user's home dir
				//String userHome = System.getProperty("user.home",".")+System.getProperty("file.separator","/");
	
				Logging.initializeLogging(userHome + logFileName,"Log File for RunWST " + new Date(), true, false);
				// redirect System.out and System.err,  rewrite any existing log
			}	
		} catch (Throwable t) {
			// should not get here
			logger.log(Level.SEVERE, "Unexpected error trying initialize log file " + logFileName, t);
		}
	}

	/**
	 *  The template extension  .st
	 */
	protected String stEnding = ".st";
	
	/**
	 *  The html extension  .html
	 */
	protected String htmlEnding = ".html";
	
	/**
	 *  The htm extension  .htm
	 */
	protected String htmEnding = ".htm";

	/**
	 *  The error listener
	 */
	protected ErrorLogListener errorListener = null;

	private static boolean stopped = false;
	// no need to synchronize this flag as it is atomic

	private Thread runThread = null;

	/**
	 *  The start delimiter  $[
	 */
	protected static String defaultStartDelim = "$[";
	
	/**
	 *  The end delimiter  ]$
	 */
	protected static String defaultEndDelim = "]$";
	
	/**
	 *  The default refresh delay  2sec.
	 */
	protected static long defaultRefreshDelay = 2;

	/**
	 *  The refresh delay
	 */
	protected long refreshDelay = defaultRefreshDelay;

	/**
	 *  Stores the last modified time for the map file
	 */
	protected Map lastModifiedTimeFileMap = new Hashtable();
	// 0 initial empty value, -1 file not found , -2 means found but not a file

	/**
	 *  Data map not set
	 */
	protected final long INITIAL_MAP_VALUE = 0;
	
	/**
	 *  Data map not found
	 */
	protected final long FILE_NOT_FOUND = -1;
	
	/**
	 *  Data map not a file
	 */
	protected final long FILE_NOT_FILE = -2;

	/**
	 *  The template to use for error message output
	 */
	protected WebStringTemplate errorTemplate = null;
	
	/**
	 *  The tree view of the data attributes
	 */
	protected AttributeTreeView attributeTreeView = null;

	
	private boolean dataMapError = true;
	// set by loadAttributes


	/**
	 *  The main method. <br>
			Usage:<br>
			RunWST template<br>
			&nbsp; &nbsp; uses inputs template.st.html template.data.html output.html<br>
			&nbsp; &nbsp;  or<br>
			RunWST template.st.html attributes.html output.html<br>
			&nbsp; &nbsp;  If output.html exists it will be renamed *.*.n before rewriting it.<br>
			&nbsp; &nbsp;  output.html will be continually updated every 2sec<br>
			&nbsp; &nbsp;    until q [Enter] or Ctrl-Z [Enter]<br>
			&nbsp; &nbsp; are typed in or the Attribute Tree display is closed.<br>
	 *
	 *@param  args  the program args
	 */
	public static void main(String args[]) {

		try {
			System.out.println();
			if ((args.length != 1) && (args.length != 3) && (args.length != 5)) {
				System.out.println("Usage:");
				System.out.println("java -jar RunWST-1.5.jar template");
				System.out.println("   uses inputs template.st.html template.data.html output.html");
				System.out.println("  or");
				System.out.println("java -jar RunWST-1.5.jar template.st data.ext page.out");
				System.out.println(" If output file exists it will be renamed *.*.n before rewriting it.");
				System.out.println(" The output file will be continually updated every 2sec");
				System.out.println("     until q [Enter] or Ctrl-Z [Enter]");
				System.out.println(" are typed in or the Attribute Tree display is closed.");
				//System.out.println(Logging.currentConfiguration());
				System.exit(1);
			}

			RunWST runTemplate = new RunWST();
			String templateFileNameArg = args[0] + ".st.html";
			if (!(new File(templateFileNameArg)).exists()) {
				// try .htm
				templateFileNameArg = args[0] + ".st.htm";
				if (!(new File(templateFileNameArg)).exists()) {
				// just leave as .html
					templateFileNameArg = args[0] + ".st.html";
				}
			}		
			String attributesFileNameArg = args[0] + ".data.html";
			if (!(new File(attributesFileNameArg)).exists()) {
				// try .htm
				attributesFileNameArg = args[0] + ".data.htm";
				if (!(new File(attributesFileNameArg)).exists()) {
				// just leave as .html
					attributesFileNameArg = args[0] + ".data.html";
				}
			}		
				
			String parentPath = (new File(args[0])).getAbsoluteFile().getParent();
			if (parentPath == null) {
				parentPath = "";
			}
			String outputFileNameArg = parentPath+File.separator+"output.html";
			String startDelimArg = defaultStartDelim;
			String endDelimArg = defaultEndDelim;

			if (args.length == 3) {
				templateFileNameArg = args[0];
				attributesFileNameArg = args[1];
				outputFileNameArg = args[2];
			}

			runTemplate.run(templateFileNameArg, attributesFileNameArg, outputFileNameArg, startDelimArg, endDelimArg,
					defaultRefreshDelay);
			// loops until stopped by user
		} catch (RunWSTException rex) {
			logger.log(Level.SEVERE, "Error in run()", rex);
		} catch (Throwable t) {
			logger.log(Level.SEVERE, "Uncaught Exception RunWST.main()", t);
		}
		//System.out.println(Logging.currentConfiguration());
		System.exit(0);
	}


	/**
	 *  Stops the program
	 */
	public void stopRun() {
		stopped = true;
		runThread.interrupt();
	}



	/**
	 *  Main processing method for the RunWST object
	 *
	 *@param  templateFileName           the template file name (with extension)
	 *@param  attributesFileName         the HTML data map file name (with extension)
	 *@param  outputFileName             the HTML output file name (with extension)
	 *@param  startDelim                 the start delimiter
	 *@param  endDelim                   the end delimiter
	 *@param  refreshDelay               the refreshDelay (in Sec), &lt; 0 is set as 0
	 *@throws  RunWSTException           if it cannot write the error to the output
	 *      file.
	 */
	public void run(String templateFileName, String attributesFileName, String outputFileName,
			String startDelim, String endDelim, long refreshDelay) throws RunWSTException {
		File canonicalTemplateFile = null;
		File canonicalAttributesFile = null;
		File canonicalOutputFile = null;
		try {
		  if (refreshDelay < 0) {
				refreshDelay = 0;
			}
			this.refreshDelay = refreshDelay;

			// get the canonical files
			try {
				canonicalOutputFile = (new File(outputFileName)).getCanonicalFile();
			} catch (IOException iex) {
				throw new RunWSTException("Error: Could not get full path for " + outputFileName, iex);
			}
			try {
				canonicalAttributesFile = (new File(attributesFileName)).getCanonicalFile();
			} catch (IOException iex) {
				throw new RunWSTException("Error: Could not get full path for " + attributesFileName, iex);
			}
			try {
				canonicalTemplateFile = (new File(templateFileName)).getCanonicalFile();
			} catch (IOException iex) {
				throw new RunWSTException("Error: Could not get full path for " + templateFileName, iex);
			}

			runThread = Thread.currentThread();
			// this is the thead that sleeps
			errorListener = new ErrorLogListener();

			WebStringTemplateGroup group = new WebStringTemplateGroup("ErrorGroup", null);
			group.setErrorListener(errorListener);
			errorTemplate = (WebStringTemplate) group.getInstanceOf("au/com/forward/runWST/error");
			if (errorTemplate == null) {
				throw new RunWSTException("Could not load template for error messages.", null);
			}
			errorTemplate.setAttribute("StringTemplateFileName", AttributeToString.htmlEscape(templateFileName));
			errorTemplate.setAttribute("AttributesFileName", AttributeToString.htmlEscape(attributesFileName));

			// check if output.html exists and rename if it does
			File parentDir = canonicalOutputFile.getParentFile();
			if (parentDir != null) {
				parentDir.mkdirs();
			}
			// skip error check here will pick it up below
			if (canonicalOutputFile.exists()) {
				if (!canonicalOutputFile.isFile()) {
					throw new RunWSTException("Error: " + outputFileName + " is not a file.", null);
				}
				File renameFile = getFileRename(outputFileName);
				if (renameFile == null) {
					throw new RunWSTException("Error: could not rename existing output file (limit 1000 backup file)." + outputFileName, null);
				}
				if (!canonicalOutputFile.renameTo(renameFile)) {
					throw new RunWSTException("Error: could not rename existing output file " + outputFileName, null);
				}
			}
			// now make our new output file
		} catch (RunWSTException rex) {
			throw rex;
		} catch (Throwable t) {
			throw new RunWSTException("Error: could not create output file " + outputFileName, t);
		}
		// OK now have output
		try {
			// if template file !exists output message html if not
			// if attributes file !exists output message html if not
			// else output waiting message

			// now go into loop until stopped
			// if attributes file changed update attributes and template and output
			// if template file changed update and output

			// continue looping until user types q Enter or presses Ctrl-Z
			ReadThread readThread = new ReadThread(System.in);
			readThread.start();
			boolean firstTimeMessage = true;
			// start reading input stream
			AttributeHashtable attributeHashtable = null;
			// not loaded yet
			while (!stopped) {		
				errorListener.clearMessages();
				// else {
				attributeHashtable = loadAttributes(attributesFileName, attributeHashtable);
        
				if (errorListener.getFirstErrorMessage() == null) {
					if (firstTimeMessage) {
						System.out.println("Loaded attributes from '" + attributesFileName + "'");
					}
					// run template
					if (!dataMapError) {
						processTemplate(attributeHashtable, templateFileName, outputFileName, startDelim, endDelim);
						if ((errorListener.getFirstErrorMessage() == null) && firstTimeMessage) {
							System.out.println(" Processed template to produce \n'" + outputFileName + "'");
						}
					}
				}

				// output any first Errors
				if (errorListener.getFirstErrorMessage() != null) {
					outputMessage(errorListener.getFirstErrorMessage(), errorListener.getFirstException(), outputFileName);
				}
				if (firstTimeMessage) {
					firstTimeMessage = false;
					System.out.println(" Template directory is "+canonicalTemplateFile.getParent());
					System.out.println("\n To quit type  q [Enter] or Ctrl-Z [Enter]");
					System.out.println(" or close the Attribute Tree display window.\n");
				}
				
				try {
					runThread.sleep(refreshDelay*1000);
				} catch (InterruptedException iex) {
				}
				
			}
		} catch (Throwable t) {
			errorListener.error("Error: Uncaught Exception found in run()", t);
			outputMessage(errorListener.getFirstErrorMessage(), errorListener.getFirstException(), outputFileName);
		}
	}


	/**
	 *  Generate the output from the template and the data
	 *
	 *@param  table             the data map
	 *@param  templateFileName  the template file name (including extension .st or .st.html)
	 *@param  delimiterStart    the start delimiter
	 *@param  delimiterEnd      the end delimiter
	 *@param  outputFileName    the output file name (including extension)
	 */
	protected void processTemplate(Map table, String templateFileName, String outputFileName, String delimiterStart, String delimiterEnd) {
		WebStringTemplate template = null;
		//WebStringTemplateGroup superGroup = null;  not used at present
		WebStringTemplateGroup group = null;
		FileWriter fw = null;

		File templateFile = null;
		File superTemplateFile = null;
		String fileName = "";

		templateFile = new File(templateFileName); // for errors
		fileName = templateFileName;
		
		try {
		// get canonical name so getParent works
			templateFile = (new File(templateFileName)).getCanonicalFile();
			// if the input file exists and it is newer and ends with .st.html
			// copy it to .st
			if (templateFile.exists() && templateFile.isFile()) {
				if (newerFile(templateFileName, lastModifiedTimeFileMap) && 
				  ( templateFileName.endsWith(stEnding + htmlEnding) || templateFileName.endsWith(stEnding + htmEnding))
					) {
					String newFileName = templateFileName.substring(0, templateFileName.lastIndexOf(htmEnding));
					try {
						// copy to .st
						copyFile(templateFileName, newFileName);
						// update templateFile with new name
						templateFile = new File(newFileName).getCanonicalFile();
					} catch (Throwable t1) {
						errorListener.error("Error copying " + templateFileName + " to " + newFileName, t1);
						return;
					}
					// update file name for template load
					fileName = newFileName;
				}
			} else {
				// does not exist or is not normal file
				Throwable st = new Throwable(templateFile.getAbsolutePath());
				errorListener.error("Error: Could not find template file " + templateFile,
						st);
				return;
			}

			// set root dir to template parent dir
			String parentPath = templateFile.getParent();
			if (parentPath == null) {
				// does not exist or is not normal file
				Throwable st = new Throwable(templateFile.getAbsolutePath());
				errorListener.error("Error: Could not find parent directory for template file " + templateFile,
						st);
				return;
			}	
			String templateName = templateFile.getName();
			group = new WebStringTemplateGroup("OutputGroup", parentPath);
			group.setErrorListener(errorListener);
			group.setMissingDataErrorPathReturn(true); // show error path returns for wep page developers.

			if (templateName.endsWith(stEnding) || templateName.endsWith(stEnding + htmlEnding)
			      || templateName.endsWith(stEnding + htmEnding) ) {
				templateName = templateName.substring(0, templateName.lastIndexOf(stEnding));
			}
			template = (WebStringTemplate) group.getInstanceOf(templateName, table);
			if (template == null) {
				errorListener.error("Error: Could not load template " + fileName, null);
				return;
			}
		} catch (Throwable t2) {
			errorListener.error("Error loading template " + fileName, t2);
			return;
		}

		/*
		 *  try {
		 *  / get canonical name so getParent works
		 *  try {
		 *  templateFile = new File(templateFileName).getCanonicalFile();
		 *  templateFileName = templateFile.getPath();
		 *  } catch (IOException ioe) {
		 *  errorListener.error("Error: Could not get full path for "+templateFileName, ioe);
		 *  return;
		 *  }
		 *  / if the input file exists and it is newer and ends with .st.html
		 *  / copy it to .st
		 *  if (templateFile.exists() && templateFile.isFile()) {
		 *  if (newerFile(templateFileName,lastModifiedTimeFileMap)  && 
		 *     ( templateFileName.endsWith(stEnding+htmlEnding) || templateFileName.endsWith(stEnding+htmEnding)) 
		 *      ) {
		 *  String newFileName = templateFileName.substring(0,templateFileName.lastIndexOf(htmEnding));
		 *  try {
		 *  / copy to .st
		 *  copyFile(templateFileName,newFileName);
		 *  / update templateFile with new name
		 *  templateFile = new File(newFileName);
		 *  } catch (Throwable t1) {
		 *  errorListener.error("Error copying "+templateFileName+" to " +newFileName, t1);
		 *  return;
		 *  }
		 *  }
		 *  / update file name for template load
		 *  fileName = templateFile.getPath();
		 *  }	else { // does not exist or is not normal file
		 *  / for now just rebuid each time
		 *  / set parent dir to parent of this template
		 *  / check if this file exists
		 *  / look in next dir up for super group
		 *  File parent = templateFile.getParentFile();
		 *  if (parent != null) {
		 *  parent = parent.getParentFile();
		 *  if (parent != null) {
		 *  superTemplateFile = new File(parent,templateFile.getName());
		 *  }
		 *  }
		 *  / if we could find the parent path
		 *  if (superTemplateFile != null) {
		 *  String superTemplateFileName = superTemplateFile.getPath();
		 *  / if the super input file exists and it is newer and ends with .st.html
		 *  / copy it to .st
		 *  if (superTemplateFile.exists() && superTemplateFile.isFile()) {
		 *  if ( newerFile(superTemplateFileName,lastModifiedTimeFileMap) &&
		 *    (  superTemplateFileName.endsWith(stEnding+htmlEnding) || superTemplateFileName.endsWith(stEnding+htmEnding)) 
		 *) {
		 *  String newSuperFileName = superTemplateFileName.substring(0,superTemplateFileName.lastIndexOf(htmEnding));
		 *  try {
		 *  / copy to .st
		 *  copyFile(superTemplateFileName,newSuperFileName);
		 *  / update superfile with new .st filename
		 *  superTemplateFile = new File(newSuperFileName);
		 *  } catch (Throwable t1) {
		 *  errorListener.error("Error copying "+superTemplateFileName+" to " +newSuperFileName, t1);
		 *  return;
		 *  }
		 *  }
		 *  / update file name for template load
		 *  fileName = superTemplateFile.getPath();
		 *  } else if (!superTemplateFile.isFile()) {
		 *  errorListener.error("Error: Super template file is not a normal file "+superTemplateFile.getPath(),null);
		 *  return;
		 *  }	else { // does not exist
		 *  Throwable st = new Throwable(superTemplateFile.getAbsolutePath());
		 *  errorListener.error("Error: Could not find template file in either the given directory or its parent directory",
		 *  st);
		 *  return;
		 *  } // else exists and is normal file
		 *  / OK have super file and no lower level file
		 *  / create super group
		 *  superGroup = new WebStringTemplateGroup("SuperGroup", superTemplateFile.getParentFile().getPath());
		 *  /superGroup.setRefreshInterval(0);
		 *  superGroup.setErrorListener(errorListener);
 			group.setMissingDataErrorPathReturn(true); // show error path returns for wep page developers.
		 *  / update file name for template load
		 *  fileName = superTemplateFile.getPath();
		 *  } else {
		 *  Throwable st = new Throwable(templateFile.getAbsolutePath());
		 *  errorListener.error("Error: Could not find template file in either the given directory or its parent directory",
		 *  st);
		 *  return;
		 *  }
		 *  } // else exists and is normal file
		 *  /  always create this group even in input file is missing and have super file.
		 *  File templateFileParent = templateFile.getParentFile();
		 *  if (templateFileParent == null) {
		 *  errorListener.error("Error: Could not find parent directory for template file " +templateFile,
		 *  null);
		 *  return;
		 *  }
		 *  group = new StringTemplateGroup("OutputGroup", templateFile.getParentFile().getPath());
		    group.setMissingDataErrorPathReturn(true); // show error path returns for wep page developers.
		 *  /group.setRefreshInterval(0);
		 *  group.setErrorListener(errorListener);
		 *  if (superGroup != null) {
		 *  / set super group if we have one
		 *  group.setSuperGroup(superGroup);
		 *  }
		 *  / strip off .st
		 *  /String fileName = templateFile.getPath();
		 *  if ( fileName.endsWith(stEnding) || fileName.endsWith(stEnding+htmlEnding)
		 *    || fileName.endsWith(stEnding+htmEnding) ) {
		 *  fileName = fileName.substring(0,fileName.lastIndexOf(stEnding));
		 *  }
		 *  if (superGroup != null) {
		 *  template = superGroup.getInstanceOf(new File(fileName).getName(),table);
		 *  } else {
		 *  template = group.getInstanceOf(new File(fileName).getName(),table);
		 *  }
		 *  if (template == null) {
		 *  errorListener.error("Error: Could not load template "+fileName, null);
		 *  return;
		 *  }
		 *  } catch (Throwable t2) {
		 *  errorListener.error("Error loading template "+fileName, t2);
		 *  return;
		 *  }
		 *  *******
		 */
		/*
		 *  **
		 *  StringTemplateTreeView viz = 	new StringTemplateTreeView("sample",template);
		 *  viz.setVisible(true);
		 */
		try {
			StringWriter stringWriterBuffer = new StringWriter(3000);
			template.write(stringWriterBuffer);
			fw = new FileWriter(new File(outputFileName));
			fw.write(stringWriterBuffer.toString());
			fw.close();
		} catch (Throwable t3) {
			errorListener.error("Error writing web page ", t3);
			try {
				fw.close();
			} catch (Throwable ct) {}
			fw = null;
		} finally {
			if (fw != null) {
				try {
					fw.close();
				} catch (Throwable ct) {
					errorListener.error("Error closing web page ", ct);
				}
				fw = null;
			}
		}
	}


	/**
	 *  Check if file is newer.
	 *
	 *@param  fileName        name of file to check
	 *@param  lastModTimeMap  map of last known modification times, keyed by canonical path
	 *@return                 true if file is newer then last know modification time
	 */
	protected boolean newerFile(String fileName, Map lastModTimeMap) {
		File file = new File(fileName);
		// find canonical name for map lookup
		String canonicalFileName = "";
		try {
			canonicalFileName = file.getCanonicalPath();
		} catch (IOException ioe) {
			errorListener.error("Error: Could not find canonical path for '" + fileName + "'", ioe);
			return true;
		}

		if (!lastModTimeMap.containsKey(canonicalFileName)) {
			lastModTimeMap.put(canonicalFileName, new Long(INITIAL_MAP_VALUE));
		}
		long lastModTime = ((Long) lastModTimeMap.get(canonicalFileName)).longValue();
		if (!file.exists()) {
			if (lastModTime != FILE_NOT_FOUND) {
				lastModTimeMap.put(canonicalFileName, new Long(FILE_NOT_FOUND));
				//lastModTime = FILE_NOT_FOUND;
				errorListener.error("Error: File '" + fileName
						 + "' does not exist. Will wait for it to appear.", null);
			}
			return true;
		}
		// else

		if (!file.isFile()) {
			if (lastModTime != FILE_NOT_FILE) {
				lastModTimeMap.put(canonicalFileName, new Long(FILE_NOT_FOUND));
				//lastModTime = FILE_NOT_FILE;
				// not a normal file
				errorListener.error("Error: File '" + fileName
						 + "' is not a normal file. Will wait for a normal file of this name to appear.", null);
			}
			return true;
		}
		// else
		if (file.lastModified() == lastModTime) {
			// no change
			return false;
		}
		// update time.
		lastModTimeMap.put(canonicalFileName, new Long(file.lastModified()));
		//lastModTime = file.lastModified();
		// else need to try and load it.
		return true;
	}



	/**
	 *  Write error message to output file using errorTemplate
	 *
	 *@param  errorMessage         the error message to write
	 *@param  t                    the exception
	 *@param  outputFileName       the name of the output file (including extension)
	 *@exception  RunWSTException  thrown if an error occures writing the output file
	 */
	protected void outputMessage(String errorMessage, Throwable t, String outputFileName)
			 throws RunWSTException {
		FileWriter fw = null;
		if (errorMessage == null) {
			return;
		}
		try {
			errorTemplate.setAttribute("FirstErrorMessage", AttributeToString.htmlEscape(errorMessage));
			errorTemplate.setAttribute("FirstException",
			      ((t==null)?"":(new WSTXmlFilter(RobustFormatter.toString(t))).toString()));

			StringWriter stringWriterBuffer = new StringWriter(3000);
			errorTemplate.write(stringWriterBuffer);
			fw = new FileWriter(new File(outputFileName));
			fw.write(stringWriterBuffer.toString());
			fw.close();
		} catch (Throwable t1) {
			try {
				fw.close();
			} catch (Throwable ct) {}
			fw = null;
			throw new RunWSTException("Error writing error message web page ", t1);
		} finally {
			if (fw != null) {
				try {
					fw.close();
				} catch (Throwable ct) {
					throw new RunWSTException("Error close error message web page ", ct);
				}
				fw = null;
			}
		}
	}


	/**
	 *  Load the HTML data map. <br>
	 *  If any error occure the first error is loaded into the result data map for display.
	 *
	 *@param  previousTable   the previous data map, if any otherwise null
	 *@param  attribFileName  the file name of the HTML data map file
	 *@return                 Description of the Return Value
	 */
	protected AttributeHashtable loadAttributes(String attribFileName, AttributeHashtable previousTable) {
		AttributeHashtable table = null;
		try {
			if (!newerFile(attribFileName, lastModifiedTimeFileMap)) {
				// no change to template or not found
				// update output if there are any messages.
				return previousTable;
			}
			String firstErrorMessage = errorListener.getFirstErrorMessage();
			if (firstErrorMessage == null) {
				// try and load attributes
				WSTAttributeReader reader = new WSTAttributeReader(errorListener);
				table = (AttributeHashtable) reader.loadDataFromFile(attribFileName);
			}
		} catch (Throwable t) {
			errorListener.error("Error loading attritbutes from '" + attribFileName + "'", t);
		}
		if (logger.isLoggable(Level.FINER)) {
			logger.finer("Resulting attributes are:\n" + AttributeToString.attributeHashtableToString(table));
		}
		
		try {
			AttributeHashtable tree = table;
			String label = "Attributes loaded from " + attribFileName;
			if (errorListener.getFirstErrorMessage() != null) {
				tree = new AttributeHashtable();
				String errMsg = errorListener.getFirstErrorMessage();
				tree.put("<FONT COLOR=RED>"
						 + AttributeToString.htmlEscape(errMsg), "Attribute map is null");
				label = "Error loading attributes from " + attribFileName;
			}
			if (attributeTreeView == null) {
				// create one
				attributeTreeView = new AttributeTreeView();
				attributeTreeView.updateAttributes(label, tree);
				attributeTreeView.addWindowListener(
					new WindowAdapter() {
						public void windowClosing(WindowEvent e) {
							//System.out.println(Logging.currentConfiguration());
							System.exit(0);
						}
					});
				attributeTreeView.setVisible(true);
			} else {
				// update it
				attributeTreeView.updateAttributes(label, tree);
			}
		} catch (Throwable t) {
			errorListener.error("Error displaying Attribute Tree view", t);
		}
		
		if (errorListener.getFirstErrorMessage() != null) {
			dataMapError = true;
			return null;
		}
		dataMapError = false;
		return table;
	}


	/**
	 *  Get an unused file name for renaming an existing file. <br>
	 *  Generated by adding .n to existing name.
	 *
	 *@param  fileName  the existing file
	 *@return           a unused new file name.
	 */
	private static File getFileRename(String fileName) {
		for (int i = 1; i < 1000; i++) {
			File f = new File(fileName + "." + i);
			if (!f.exists()) {
				return f;
			}
		}
		return null;
	}



	/**
	 *  Copies the from file to the to file.
	 *
	 *@param  fromFileName     the name of the file to copy from
	 *@param  toFileName       the name of the file to copy to
	 *@return                  true if copy performed else false if fromFile is the
	 *      same as toFile or some error occured
	 *@exception  IOException  thrown if error occured while copying
	 *@exception  IllegalArgumentException  thrown if to file or from file names are empty.
	 *@exception  NullPointerException  thrown if to file or from file names are null.
	 */
	public static boolean copyFile(String fromFileName, String toFileName) throws IOException {
		if (fromFileName == null) {
			throw new NullPointerException(
					"File name to copy from cannot be null");
		}

		fromFileName = fromFileName.trim();

		if (fromFileName.length() == 0) {
			throw new IllegalArgumentException(
					"File name to copy from cannot be empty");
		}

		if (toFileName == null) {
			throw new NullPointerException(
					"File name to copy to cannot be null");
		}

		toFileName = toFileName.trim();

		if (toFileName.length() == 0) {
			throw new IllegalArgumentException(
					"File name to copy to cannot be empty");
		}
		// are these the same file
		// check the canonical names
		File toFile = (new File(toFileName)).getCanonicalFile();
		File fromFile = (new File(fromFileName)).getCanonicalFile();
		if (toFile.compareTo(fromFile) == 0) {
			// nothing to do as files are the same
			return false;
		}

		FileInputStream fis = new FileInputStream(fromFileName);
		FileOutputStream fos = new FileOutputStream(toFileName);
		FileChannel fcin = fis.getChannel();
		FileChannel fcout = fos.getChannel();

		// do the file copy
		fcin.transferTo(0, fcin.size(), fcout);

		// finish up
		fcin.close();
		fcout.close();
		fis.close();
		fos.close();

		logger.finer(" Copied:" + fromFileName + " to "
				 + toFileName);
		return true;
	}




	/*
	 *  Typical main for these threads uses www.forward.com.au  ThreadReturn package
	 *  public static void main(String args[]) {
	 *  T t = new T();
	 *  try {
	 *  System.out.println("Enter input (user ctrl+Z Enter to terminate):");
	 *  t.start();
	 *  / uncomment these lines to interrupt input after 5 seconds
	 *  /sleep(5000);
	 *  /System.out.println("Interrupt t now");
	 *  /t.interrupt();
	 *  /
	 *  String threadResult = (String)ThreadReturn.join(t);
	 *  System.out.println("Thread returned \n"+threadResult);
	 *  } catch (InterruptedException ex) {
	 *  / some error running t, handle it here
	 *  if (ex instanceof ThreadException) {
	 *  / t throw a Throwable which is the cause of this ThreadException
	 *  System.err.println(" Thread throw "+ex.toString());
	 *  } else {
	 *  / ThreadReturn.join(t1) was interrupted
	 *  System.err.println(" Thread was interrupted "+ex.toString());
	 *  }
	 *  }
	 *  }
	 *  protected class InputCheckingThread extends Thread {
	 *  private InputStream in = System.in;
	 *  public InputCheckingThread() {
	 *  setDaemon(true); // don't wait for this thread is we want to stop
	 *  }
	 *  public void run(){
	 *  ReadThread readThread = null;
	 *  String lines = "";
	 *  try {
	 *  readThread = new ReadThread(in);
	 *  readThread.start(); // start reading input stream
	 *  while (!readThread.eof()) {
	 *  if (readThread.available()<=0) {
	 *  sleep(1000); // wait for some input
	 *  }
	 *  lines += readThread.readString();
	 *  ThreadReturn.ifInterruptedStop();
	 *  }
	 *  ThreadReturn.save(lines);
	 *  } catch (Throwable tex) {
	 *  ThreadReturn.save(tex);
	 *  }
	 *  }
	 */
	/*
	 *  here we only need this part
	 *  and not all of these methods are used either
	 */
	/**
	 *  Handles checking the console for the "quit" character 'q'
	 *
	 *@author     matthew ford
	 *@created    November 28, 2003
	 */
	protected class ReadThread extends Thread {
		private int bufferSize = 1024;
		private char[] inBuffer = new char[bufferSize];
		private int charsAvailable = 0;
		private int charsRead = 0;
		private char[] outBuffer = new char[bufferSize];
		private InputStream input = null;
		private Throwable errorOnStream = null;
		private boolean closed = false;
		private boolean endOfFile = false;
		private InputStreamReader reader = null;


		/**
		 *  Constructor for the ReadThread object
		 *
		 *@param  in  input stream, usually System.in
		 */
		public ReadThread(InputStream in) {
			reader = new InputStreamReader(in);
			setDaemon(true);
			// stop this thread when the program stops
		}


		/**
		 *  Main processing method for the ReadThread object
		 */
		public void run() {
			try {
				while (true) {
					// block here waiting for input
					charsRead = reader.read(inBuffer, 0, bufferSize);
					if (charsRead == -1) {
						endOfFile = true;
						return;
					}
					while (available() != 0) {
						sleep(100);
					}
					bufferChars();
					// read them
					// normally read by some other thread
					String input = readString();
					// look for 'q'
					if ((input.indexOf('q') >= 0) || (input.indexOf('Q') >= 0)) {
						return;
					}
				}
			} catch (Throwable t) {
				setErrorOnStream(t);
			} finally {
				stopRun();
				// stop the application when this thread returns for any reason
			}
		}


		/**
		 *  How many characters are available to read
		 *
		 *@return    number of characters available to read
		 */
		public synchronized int available() {
			return charsAvailable;
		}


		/**
		 *  Has end of file been reached
		 *
		 *@return    true if end of file
		 */
		public synchronized boolean eof() {
			return ((charsAvailable == 0) && endOfFile);
		}


		/**
		 *  Sets the errorOnStream 
		 *
		 *@param  t  the exception
		 */
		public synchronized void setErrorOnStream(Throwable t) {
			errorOnStream = t;
		}


		/**
		 *  Is there an error on the stream
		 *
		 *@return    null for no error, else the exception
		 */
		public synchronized Throwable streamError() {
			if (charsAvailable == 0) {
				return errorOnStream;
			}
			//
			return null;
			// no errors
		}


		/**
		 *  Read the currently available characters, non-blocking. <br>
		 *  Returns an empty string if no characters available.
		 *
		 *@return    currently available characters, empty string if no characters available
		 */
		public synchronized String readString() {
			String rtn = new String(outBuffer, 0, charsAvailable);
			charsAvailable = 0;
			return rtn;
		}


		/**
		 *  Transfer the character read from the input to the available character buffer ready for a non-blocking read.
		 */
		private synchronized void bufferChars() {
			if (charsAvailable == 0) {
				System.arraycopy(inBuffer, 0, outBuffer, 0, charsRead);
				charsAvailable = charsRead;
				charsRead = 0;
			}
			// else wait until previous characters are taken
		}

	}

}

