/*
 *  Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *  copyright notice, this list of conditions and the following disclaimer
 *  in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *  include, in the documentation and/or other materials provided with the
 *  distribution, a statement that the software has been modified and
 *  identify the person who was responsible for the modifications.
 *  3. Neither the name of the author nor Forward Computing and Control may be
 *  used to endorse or promote products derived from this software
 *  without specific prior written permission.
 *  4.  All advertising materials mentioning features or use of this
 *  software must display the following acknowledgment:
 *  "This product includes software developed by Matthew Ford
 *  and Forward Computing and Control Pty. Ltd.
 *  (http://www.forward.com.au)."
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package au.com.forward.webstringtemplate.filters;
import au.com.forward.runWST.wstAttributeReader.WSTAttributeReader;
import au.com.forward.runWST.maps.*;
import java.util.logging.*;
import au.com.forward.logging.*;

/**
 *  This is a plain filter, it does not change the result of the object's toString(). <br>
 *  This is the default implementation of {@link IWSTFilter}.  You would normally extend from
 * this class.  See {@link WSTXmlFilter} for an example.
 *
 *@author     matthew ford
 *@created    December 28, 2003
 */
public class WSTPlainFilter implements IWSTFilter {

	/**
	 *  The Java logger for this class
	 */
	public final static Logger logger = Logger.getLogger("au.com.forward.webstringtemplate.filters.WSTPlainFilter");

	/**
	 *  the object being filtered
	 */
	protected Object obj = null;
	private String path = "";  // default to no path.


	/**
	 *  Default constructor for the WSTPlainFilter object
	 */
	public WSTPlainFilter() { }


	/**
	 *  Constructor for the WSTPlainFilter object
	 *
	 *@param  obj   the object to filter, can be null
	 */
	public WSTPlainFilter(Object obj) {
		setObject(obj);
	}

	/**
	 *  Constructor for the WSTPlainFilter object
	 *
	 *@param  obj   the object to filter, can be null
	 */
	public WSTPlainFilter(Object[] obj) {
		setObject(obj);
	}

	/**
	 *  returns the object being filtered, null if not set.
	 *
	 *@return    the object beinig filtered.
	 */
	public Object getObject() {
		return obj;
	}


	/**
	 *  sets the object being filtered.
	 *
	 *@param  obj  the object to be filtered
	 *@throws IllegalStateException if the object is this filter itself or if it wraps this filter.
	 */
	public final void setObject(Object obj) {
		
		// logging output
		Level level = Level.FINER;
		if (logger.isLoggable(level)) {
			String msg = "";
			Object trueObj = obj;
			while (trueObj instanceof IWSTFilter) {
				msg += " " + trueObj.getClass().getName() + " containing ";
				trueObj = ((IWSTFilter) trueObj).getObject();
			}
			msg += "" + trueObj;
			logger.log(level, msg);
		}

		// checking for circular reference
		Object trueObj = obj;
		while (trueObj instanceof IWSTFilter) {
			if (trueObj == this) {
				throw new IllegalStateException("Circular reference: Tried to add yourself as the object of this filter");
			}
			trueObj = ((IWSTFilter) trueObj).getObject();
		}	
		this.obj = obj;
	}


	/**
	 *  returns the filtered string of this object or "" (empty string) if object null
	 *
	 *@return    the filtered string of this object. If obj.toString() throws an exception then the exception message is returned.
	 */
	public String toString() {
		if (obj instanceof IWSTFilter) {
			// use that filter to handle the object it holds.
			return obj.toString();
		}
		String rtn = convertToString(obj);
		if (this instanceof WSTPlainFilter) {
			logger.exiting(this.getClass().getName(), "toString", rtn);
		}

		return rtn;
	}


	/**
	 *  returns the attribute name path of the object in this filter, or "" if not set
	 *
	 *@return    The path value or "" if not set
	 */
	public String getPath() {
		return path;
	}


	/**
	 *  sets the attribute name path of the object in this fiter, "" if not set
	 *
	 *@param  newPath                    The new path value, stores "" for null
	 *@throws  IllegalArgumentException  if not null and not a valid attribute name
	 */
	public void setPath(String newPath) {
		if (newPath == null) {
			newPath = "";
		}
		if ( !newPath.equals("") ) {
			// only these chars are allowed by the stringtemplate parser
			if (!WSTAttributeReader.attributeNamePattern.matcher(newPath).matches()) {
				throw new IllegalArgumentException("Attribute name must be letter or _ followed by letter _ or digit and separated by dots.  Found '" + newPath + "'");
			}
		}
		path = newPath;
	}




	/**
	 *  Returns a new instance of this filter. <br>
	 *  The object and path themselves are not copied. This usually correct as the object will be
	 *  reset and the path. Any other other objects in the filter are need to be copied across
	 *  in the over of this method.
	 *
	 *@return    a new instance of this filter
	 */
	public IWSTFilter newInstance() {
		try {
			return (IWSTFilter)getClass().getConstructor(null).newInstance(null);
		} catch (Throwable t) {
			throw new RuntimeException("Error: getting newInstance of "+getClass().getName(),t);
		}
	}


	/**
	 *  Get the string representation of this object.<br>
	 *  If an error occures calling <code>obj.toString()</code>, return the error message as the string.
	 *
	 *@param  trueObj  the object to convert to a string
	 *@return          the string represention of the object or an error message.  This method does not throw exceptions.
	 */
	public final static String convertToString(Object trueObj) {
		String rtn = "";
		if (trueObj != null) {
			// if null leave rtn as ""
			// try toString()
			try {
				rtn = "" + trueObj;
			} catch (Throwable t) {
				logger.log(Level.SEVERE, "Error converting " + trueObj.getClass().getName() + " to String", t);
				rtn = ("Error converting " + trueObj.getClass().getName() + " to String\n" + RobustFormatter.toString(t));
				//throw new RuntimeException("Error calling toString() on "+obj.getClass().getName(),t);
			}
		}

		return rtn.trim();
	}

	
}

